/*
 * receive.c
 * 
 * Takes as input signals from microphone/tone decoder circuit to construct a
 * key scan code, then outputs the corresponding characters to the LCD
 */

/************************ Lab 5 directives and constants ***********************/
#include "trtSettings.h"
#include "trtkernel_1284.c"
#include <stdio.h>
#include <avr/sleep.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h> // needed for lcd_lib
#include "keymap.h"     // contains character arrays for decoding keys
#include <math.h>

int args[3] ;

// Variables for circular buffer of keyboard scan codes
volatile int lcdWR_pnt;             // pointer for keyboard write to lcd circular buffer
volatile int lcdRD_pnt;             // pointer for lcd read
volatile uint16_t lcd_buffer[16];   // array for lcd circular buffer (only contains chars)
#define LCD_BUFFER_RDY 3            // semaphore for lcd_buffer
int lcd_buffer_size = 16;           // size of lcd buffer (for % operations)

// Variable to contain keyboard scan code
volatile uint16_t kbd_scan_code;

// Variables for handling keyboard scan codes
uint8_t started;		// flag to signal beginning of packet
uint8_t bit_count;		// variable to keep track of bit position in key scan code
uint8_t parity;			// variable for parity check

volatile uint8_t reset; // flag to signal timeout

/**************** Serial communication directives and constants ****************/
// Serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

/*************************** LCD directives and constants ********************************************/
#include "lcd_lib.h"

const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";;
const int8_t LCD_clearline[] PROGMEM = "_               \0";;

int8_t lcd_display_buffer[17];	// LCD display buffer

/****************************************************************************** 
* INT2_vect ISR <Triggers on falling edge of PINB2(8 kHz tone decoder output)> 
*------------------------------------------------------------------------------
*    Reads in data from PINB0(5 kHz tone decoder output) bit by bit to 
*    construct a key scan code which is then stored in a circular read buffer 
*    provided that there is space in the buffer. Also does checks to ensure 
*    data is valid.
*    
*    Transmission protocol is very similar to PS/2 protocol
*      - An alternating high frequency (8 kHz) tone represents CLK
*      - An alternating low frequency (5 kHz) tone represents DATA
*      - Presence of a frequency corresponds to logic 0, as tone decoder circuit
*        is active low
*      - On falling edge of CLK, DATA represents current bit
*	   - Each keystroke packet is sent as 12 bits: 
*        +-+--------+-+-+-+      s = start bit
*        |s|  scan  |h|p|f|   scan = 8-bit key scan code
*        +-+--------+-+-+-+      h = shift flag
*         0   1:8   9 10 11      p = parity bit (odd)
*                                f = stop bit
*      - CLK toggles every 25 mSec, DATA updates every 50 mSec, corresponding
*        to a total transmission time of (600 mSec per packet) followed by a 
*        100 mSec pause
*/
ISR (INT2_vect)
{	
	// Make sure clock line is low, if not ignore this transition
  	if(PINB & (1<<PINB2)){
    	return;
  	}
	
	// If time elapsed between receiving two bits is over a certain 
	// threshold, start over!
	if (reset) {
		fprintf(stdout,"reset!\n\r");
		started = 0;
	}	
	
	// Print out current bit value <REMOVE>
	//fprintf(stdout,"%d\n\r", (PINB & (1<<PINB0)));

	// Read in DATA bits
	if (!started) {							
		// DATA[0] = Start bit (should always be 0)
		//   if == 0 -> Initialize variables and begin constructing key scan code 
		//      else -> Do nothing
		if ((PINB & (1<<PINB0)) == 0x00) {
			started = 1;
			bit_count = 0;
			kbd_scan_code = 0;
			parity = 0;
			reset = 0; 
			TCNT3 = 0; 
		}
		return;
	} else if (bit_count < 9) { 			
		// DATA[1:9] = bits of key scan code (least significant bit first)
		//   if == 1 -> Add 1 in appropriate position in key scan code, xor 
		//              parity for parity check, increment bit_count
		//      else -> Only increment bit_count
		if (PINB & (1<<PINB0)) {
			kbd_scan_code |= (1<<bit_count);
			parity ^= 1;
		}
		bit_count++;
		reset = 0; 
		TCNT3 = 0; 
		return;
	} else if(bit_count == 9) { 			
		// DATA[10] = odd parity bit (should always make # of 1's in packet odd)
		//   if odd number of 1's -> Increment bit_count
		//                   else -> Start over!
		if ((PINB & (1<<PINB0)) ^ parity) {
			bit_count++;
			reset = 0; 
			TCNT3 = 0; 
		} else {
			fprintf(stdout,"incorrect parity bit!\n\r");
			started = 0;	
		}
		return;
	} else {								
		// DATA[11] = stop bit (Should always be 1)
		//   if == 1 -> Proceed to key scan code handling block
		//      else -> Start over!
		if ((PINB & (1<<PINB0)) == 0x00) {
			started = 0;
			fprintf(stdout,"low stop bit!\n\r");
			return;
		}
		started = 0;
		bit_count = 0;
		reset = 0; 
		TCNT3 = 0; 
	}
	
	// Print out key scan code <REMOVE>
	//fprintf(stdout,"kbd=%x\n\r", kbd_scan_code);
	
	// If write pointer is right before read pointer, buffer is full.
	// In this case, simply drop the key scan code (HOPEFULLY RARE CASE)
	if ((lcdWR_pnt + 1) % lcd_buffer_size != lcdRD_pnt) {
		lcd_buffer[lcdWR_pnt] = kbd_scan_code;
		lcdWR_pnt = (lcdWR_pnt + 1) % lcd_buffer_size;
		trtSignal(LCD_BUFFER_RDY);
	}
}

/****************************************************************************** 
* TIMER3_COMPA_vect ISR <Triggers after 80 mSec> 
*------------------------------------------------------------------------------
*    Simply sets a flag high to indicate that more than 80 mSec has elapsed
*    since the last time TCNT3 was reset. Used to detect when the time elapsed
*    between successive bits in a packet is too high (most likely due to a 
*    dropped bit)
*/
ISR (TIMER3_COMPA_vect)
{
	reset = 1;
}

/****************************************************************************** 
* keyScanCode2Char()
*------------------------------------------------------------------------------
*    Takes as input an 8-bit key scan code and a shift flag and outputs 
*    the corresponding character. This is done simply by using the the 
*    key scan code to index into one of two arrays (depending on whether 
*    shift is high) and returning the corresponding char. The two arrays 
*    can be found in keymap.h
*/
char keyScanCode2Char(uint8_t data, uint8_t shift){
 	char to_ret; 
  	if(shift){
		to_ret = pgm_read_byte(&(shiftmap[data])); // Grab char from shift array
  	} else {
		to_ret = pgm_read_byte(&(keymap[data]));   // Grab char from key array
	}
  	return to_ret;
}

/****************************************************************************** 
* lcdReadKeyScanCode()
*------------------------------------------------------------------------------
*    Reads and returns a key scan code from the lcd_buffer. Waits on a 
*    semaphore if the lcd_buffer is currently empty
*/
uint16_t lcdReadKeyScanCode(){
	uint16_t to_ret;
	
	// Wait for new key scan code to arrive
	trtWait(LCD_BUFFER_RDY);
	
	// Read key scan code from lcd_buffer, increment read pointer
	to_ret = lcd_buffer[lcdRD_pnt];
	lcdRD_pnt = (lcdRD_pnt + 1) % lcd_buffer_size;
  	return to_ret;
}

/****************************************************************************** 
* lcd_display()
*------------------------------------------------------------------------------
*    Responsible for updating the LCD display. Does so by constantly checking
*    the circular buffer for new key scan codes, checks for special cases and 
*    displays the corresponding characters to the LCD as long as they are valid
*/
void lcd_display(void* args) 
{
	// Initialize variables
  	uint32_t rel, dead ;    // release/deadlines for task management
	uint16_t full_code = 0; // [1 bit shift_flag|8 bit key_scan_code]
	uint8_t key_code = 0;   // key scan code
	uint8_t shift = 0;      // flag to indicate whether shift key is pressed
	int lcdX, lcdY;         // (x,y) coordinates for LCD (initialized to [0,0])
	lcdX = lcdY = 0;		
	
	// Initially LCD has _ at (0,0) position
	CopyStringtoLCD(LCD_clearline, lcdX, lcdY);

	while(1)
	{
		// Get key scan code from circular buffer
		full_code = lcdReadKeyScanCode();
		key_code = full_code;
		shift = full_code >> 8;
		if (key_code == 0x76) { 		
			// <escape> key pressed, clear LCD
			lcdX = lcdY = 0;
			LCDclr();
			CopyStringtoLCD(LCD_clearline, lcdX, lcdY);
			LCDGotoXY(lcdX,lcdY);
		} else if (key_code == 0x5a) { 	
		  	// <enter> key, clear line
			CopyStringtoLCD(LCD_clearline, 0, lcdY);
			lcdX = 0;
			LCDGotoXY(lcdX,lcdY);
		} else if (key_code == 0x66) { 	
			// <backspace> key, move cursor back one position
			sprintf(lcd_display_buffer, " ");
			LCDGotoXY(lcdX, lcdY);
  			LCDstring(lcd_display_buffer, strlen(lcd_display_buffer));
			lcdX--;
			// Handle wraparound case
			if (lcdX < 0) {
				lcdX = 15;
				lcdY = !lcdY;
			}
			sprintf(lcd_display_buffer, "_");
			LCDGotoXY(lcdX, lcdY);
  			LCDstring(lcd_display_buffer, strlen(lcd_display_buffer));
		} else {
			// Get character corresponding to key scan code
			char c = keyScanCode2Char(key_code, shift);
			
			// Only output character to LCD if it is valid
			if (c != '?' || key_code == 0x4a) {
				sprintf(lcd_display_buffer, "%c", c);
				LCDGotoXY(lcdX, lcdY);
  				LCDstring(lcd_display_buffer, strlen(lcd_display_buffer));
				lcdX++;
				
				// Handle wraparound case
				if (lcdX > 15) {
					lcdX = 0;
					lcdY = !lcdY;
				}
				sprintf(lcd_display_buffer, "_");
				LCDGotoXY(lcdX, lcdY);
  				LCDstring(lcd_display_buffer, strlen(lcd_display_buffer));
			}
			// Print out variables for debugging <REMOVE>
			//fprintf(stdout,"c=%x, s=%x, k=%c, (x,y)=(%d, %d)\n\r", key_code, shift, keyScanCode2Char(key_code, shift), lcdX, lcdY);	
		}
	    rel = trtCurrentTime() + SECONDS2TICKS(0.05);
	    dead = trtCurrentTime() + SECONDS2TICKS(0.20);
	    trtSleepUntil(rel, dead);
	}
}

/****************************************************************************** 
* init_lcd()
*------------------------------------------------------------------------------
*    Initializes LCD
*/
void init_lcd(void) 
{
	LCDinit();                              // Initialize the display
	LCDcursorOFF();
	LCDclr();                               // Clear the display
	LCDGotoXY(0,0);	
	CopyStringtoLCD(LCD_initialize, 0, 0);
}

/****************************************************************************** 
* initialize()
*------------------------------------------------------------------------------
*    Initialize everything
*/
void initialize()
{
	// Turn on onboard LED
	DDRD = 0x04;
	PORTD = PORTD ^(0<<PORTD2);

	// Set up INT2 interrupt
	EICRA = (1<<ISC21);     // set INT2 to interrupt on falling edge
	EIMSK = (1<<INT2);  	// enable interrupt by INT2
	
	// Make B.0 and B.2 inputs
	DDRB = 0x00;

	// Set up Timer3
	TCCR3B= 3;              // set prescalar to divide by 64
	TCCR3B |= (1<<WGM32)    // Set Timer3 to CTC mode
	TIMSK3= (1<< OCIE3A);   // turn on Timer3 Output Compare A Match interrupt
	OCR3A = 19999 ;         // Set interrupt to trigger after 80 mSec

  	// Initialize the LCD
  	init_lcd();
  	LCDclr();

	// Crank up the ISRs
	sei();
}

/****************************************************************************** 
* main()
*------------------------------------------------------------------------------
*    Entry point and task scheduler loop
*/
int main(void)
{
	// Run initialization stuff
	initialize();

	// Initialize UART (trt_uart_init() is in trtUart.c)
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"\n\r Keyboard Reading Started...\n\r\n\r");

	// Start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack

	// Create Semaphores (first two are necessary for uart)
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; 	// uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  	// user typed <enter>
	trtCreateSemaphore(LCD_BUFFER_RDY, 0) ; 	// new key scan code in LCD buffer

	// Create task
	trtCreateTask(lcd_display, 400, SECONDS2TICKS(0.01), SECONDS2TICKS(0.3),&(args[1]));

	// Initialize variables
	started = 0;
	kbd_scan_code = 0;
	bit_count = 0;
	parity = 0;
	lcdWR_pnt = 0;
	lcdRD_pnt = 0;

	// Idle Task - just sleeps the cpu to save power every time it executes
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	while (1) {
		sleep_cpu();
	}
}


