/*
 * transmit.c
 * 
 * Takes as input signals resulting from PS/2 keyboard presses and outputs the 
 * corresponding dual-tones on the speakers
 */

/************************ Lab 5 directives and constants ***********************/
#include "trtSettings.h"
#include "trtkernel_1284.c"
#include <stdio.h>
#include <avr/sleep.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h> // needed for lcd_lib
#include "keymap.h"     // contains character arrays for decoding keys
#include <math.h>

int args[3] ;

// Variables for circular buffer of keyboard scan codes
volatile int transWR_pnt;           // pointer for keyboard write to transmitter circular buffer
volatile int transRD_pnt;           // pointer for transmitter read
volatile uint16_t trans_buffer[32];	// array for transmitter circular buffer
int trans_buffer_size = 32;         // size of transmitter buffer (for % operations)
#define TRANS_BUFFER_RDY 4          // semaphore for trans_buffer

// Variable to contain keyboard scan code
volatile uint16_t kbd_scan_code;

// Variables for handling keyboard scan codes
uint8_t started;		// flag to signal beginning of packet
uint8_t bit_count;		// variable to keep track of bit position in key scan code
uint8_t parity;			// variable for parity check
uint8_t shift;			// flag to signal if shift key is held
uint8_t extended;		// flag to signal if extended key was pressed
uint8_t release;		// flag to signal if a key was released

/**************** Serial communication directives and constants ****************/
// Serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 		// user hit <enter>
#include "trtUart.h"
#include "trtUart.c"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

/*************************** DDS directives and constants ********************************************/
// DDS variables 
unsigned int Fout1, Fout2;                          // output frequency
volatile unsigned long increment1, increment2;      // determines how quickly we step through sineTable 
volatile unsigned long accumulator1, accumulator2;  // keeps track of where we are in the sineTable
volatile unsigned char sinIdx1, sinIdx2;            // index into sineTable 
signed char sineTable[256];                         // array of samples used to contruct a sine wave
volatile unsigned int rampIdx;                      // index into rampTable 					
char rampTable[256];                                // array of coefficents for ramping up and down of tone
#define RAMP_CAP 255                                // ~4 msec ramping up/down time 
volatile char ramp;                                 // flag to initiate ramping up or down

// Time variables
// the volatile is needed because the time is only set in the ISR
#define countMS 62                  // ~62 interrupts correspond to 1 msec
volatile unsigned int time;         // counts mSec
volatile char count;                // variable to count interrupts (~62 = 1 msec)

// Index for sine table build
unsigned int i;

/****************************************************************************** 
* INT2_vect ISR <Triggers on falling edge of PINB2(Keyboard CLK)> 
*------------------------------------------------------------------------------
*    Reads in data from PINB0(Keyboard DATA) bit by bit to construct a key 
*    scan code which is then stored in a circular read buffer provided that
*    there is space in the buffer. Also does checks to ensure data is valid.
*    
*    PS/2 Protocol:
*	   - On falling edge of clock line, data line represents current bit
*	   - Each keystroke packet is sent as 11 bits: 
*        +-+--------+-+-+	   s = start bit
*        |s|  scan  |p|f|	scan = 8-bit key scan code
*        +-+--------+-+-+      p = parity bit (odd)
*         0   1:8    9 10      f = stop bit
*      -   On pushing a button -> corresponding key scan code sent once
*      -   On holding a button -> corresponding key scan code sent repeatedly
*      - On releasing a button -> release code sent once, then corresponding 
*                                 key scan code sent once
*/
ISR (INT2_vect)
{	
	// Make sure clock line is low, if not ignore this transition
  	if(PINB & (1<<PINB2)){
    	return;
  	}
	
	// Read in DATA bits
	if (!started) {							
		// DATA[0] = Start bit (should always be 0)
		//   if == 0 -> Initialize variables and begin constructing key scan code 
		//      else -> Do nothing
		if ((PINB & (1<<PINB0)) == 0x00) {
			started = 1;
			bit_count = 0;
			kbd_scan_code = 0;
			parity = 0; 
		}
		return;
	} else if (bit_count < 8) { 			
		// DATA[1:8] = bits of key scan code (least significant bit first)
		//   if == 1 -> Add 1 in appropriate position in key scan code, xor 
		//              parity for parity check, increment bit_count
		//      else -> Only increment bit_count
		if (PINB & (1<<PINB0)) {
			kbd_scan_code |= (1<<bit_count);
			parity ^= 1;
		}
		bit_count++;
		return;
	} else if(bit_count == 8) { 			
		// DATA[9] = odd parity bit (should always make # of 1's in packet odd)
		//   if odd number of 1's -> Increment bit_count
		//                   else -> Start over!
		if ((PINB & (1<<PINB0)) ^ parity) {
			bit_count++;
		} else {
			//fprintf(stdout,"incorrect parity bit!\n\r");
			started = 0;	
		}
		return;
	} else {								
		// DATA[10] = stop bit (Should always be 1)
		//   if == 1 -> Proceed to key scan code handling block
		//      else -> Start over!
		if ((PINB & (1<<PINB0)) == 0x00) {
			started = 0;
			//fprintf(stdout,"low stop bit!\n\r");
			return;
		}
		started = 0;
		bit_count = 0;
	}
	// Handle resulting key scan code
	if (kbd_scan_code == 0xF0) {				
		// Handle release code
		//   Raise release flag
		release = 1;
		kbd_scan_code = 0;
		return;
	} else if (kbd_scan_code == 0x12 || kbd_scan_code == 0x59) { 
		// Handle Shift keys 
		//   Raise or lower one of two shift flags (needed for 2 shift keys)
		if (release) {
			if (kbd_scan_code == 0x12) shift &= ~0x01;
			else shift &= ~0x02;
			release = 0;
		} else {
			if (kbd_scan_code == 0x12) shift |= 0x01;
			else shift |= 0x02;
		}
		return;
	} else if (kbd_scan_code == 0xe0) { 
		// Handle extended keys
		//   Raise or lower extended flag 
		if (release) {
			extended = 0;
		} else {
			extended = 1;
		}
		return;
	} else {
		// Handle all other keys
		//   if key released (release flag high) -> set release flag low
		//                                  else -> save key scan code to lcd_buffer
		if (release) {
			release = 0;
		} else {
			// If write pointer is right before read pointer, buffer is full.
			// In this case, simply drop the key scan code (HOPEFULLY RARE CASE)
			// Save key scan code to trans_buffer as long as there is space
			if ((transWR_pnt + 1) % trans_buffer_size != transRD_pnt) {
				// If shift is pressed, set corresponding bit high
				if (shift) kbd_scan_code |= (1 << 8);
				
				// Save key scan code to buffer
				trans_buffer[transWR_pnt] = kbd_scan_code;
				transWR_pnt = (transWR_pnt + 1) % trans_buffer_size;
				
				// Signal to transmitter that new key scan code is available
				trtSignal(TRANS_BUFFER_RDY);
			}
		}
	}
}

/****************************************************************************** 
* Timer0_OVF_vect ISR <Triggers on Timer 0 Overflow> 
*------------------------------------------------------------------------------
*    Generates a sinewave using pulse-width modulation. Also acts as a 
*	 mSec timer. 
*/
ISR (TIMER0_OVF_vect)
{ 
	// Increment accumulators as determined by Fout1 and Fout2
	accumulator1 += increment1;
	accumulator2 += increment2;
	
	// Get indexes into sineTable from accumulators
	sinIdx1 = (char)(accumulator1 >> 24);
	sinIdx2 = (char)(accumulator2 >> 24);
	
	// Set new duty cycle
	OCR0A = 128 + (((sineTable[sinIdx1]+sineTable[sinIdx2]) * rampTable[rampIdx])>>7);
	
	// Ramp up/down depending on status of ramp flag
	if (ramp) {
		if (rampIdx < RAMP_CAP) rampIdx++;
	} else {
		if (rampIdx > 0) rampIdx--;
	}

	// Decrements mSec time variable
	count--;
	if (count <= 0) {
		count = countMS;
		if (time > 0) --time;
	}
}

/****************************************************************************** 
* transReadKeyScanCode()
*------------------------------------------------------------------------------
*    Reads and returns a key scan code from the trans_buffer. Waits on a 
*    semaphore if the trans_buffer is currently empty
*/
uint16_t transReadKeyScanCode(){
	uint16_t to_ret;
	
	// Wait for new key scan code to arrive
	trtWait(TRANS_BUFFER_RDY);	
	
	// Read key scan code from trans_buffer, increment read pointer
	to_ret = trans_buffer[transRD_pnt];
	transRD_pnt = (transRD_pnt + 1) % trans_buffer_size;
  	return to_ret;
}

/****************************************************************************** 
* transmitter()
*------------------------------------------------------------------------------
*    Responsible for transmitting key scan codes. Does so by constantly checking
*    the transmitter circular buffer for new key scan codes, outputting new key
*	 scan codes as they arrive
*
*	 Transmission protocol is very similar to PS/2 protocol
*      - An alternating high frequency (8 kHz) tone represents CLK
*      - An alternating low frequency (5 kHz) tone represents DATA
*      - Presence of a frequency corresponds to logic 0, as tone decoder circuit
*        is active low
*      - On falling edge of CLK, DATA represents current bit
*	   - Each keystroke packet is sent as 12 bits: 
*        +-+--------+-+-+-+      s = start bit
*        |s|  scan  |h|p|f|   scan = 8-bit key scan code
*        +-+--------+-+-+-+      h = shift flag
*         0   1:8   9 10 11      p = parity bit (odd)
*                                f = stop bit
*      - CLK toggles every 25 mSec, DATA updates every 50 mSec, corresponding
*        to a total transmission time of 600 mSec per packet followed by a 
*        100 mSec pause
*/
void transmitter(void* args) 
{
	// Initialize variables
  	uint32_t rel, dead ;	// release/deadlines for task management
	uint16_t key_code = 0;	// key scan code
	uint8_t t_parity = 0;	// parity variable for error checking
	unsigned int hi = 8000;	// high-frequency tone (corresponds to CLK 0)
	unsigned int lo = 5000; // low-frequency tone  (corresponds to DATA 0)
	unsigned int off = 0;   // no tone (corresponds to CLK/DATA 1)

	while(1)
	{
		// Get key scan code from circular buffer
		key_code = transReadKeyScanCode();
		
		// Ramp up
		ramp = 1;
		rampIdx = 0;
		
		// Initialize variables
		t_parity = 0;
		Fout1 = Fout2 = off;
		
		// Drive CLK and Data lines
		for(int b=0; b<24; b++ ) {
			if (b < 2) {	
				// DATA[0] = Start bit (should always be 0)
				Fout1 = lo;
			}
			else if (b < 20) {
				// DATA[1:8] = bits of key scan code (least significant bit first)
				//   DATA[9] = shift flag  
				if (key_code & (1<<(b/2-1))) {
					Fout1 = off;
					t_parity ^= b%2;
				} else {
					Fout1 = lo;
				}
			}
			else if (b < 22) {
				// DATA[10] = odd parity bit (should always make # of 1's in packet odd)
				if (t_parity == 0) {
					Fout1 = off;	
				} else {
					Fout1 = lo;
				}
			} else {
				// DATA[11] = stop bit (Should always be 1)
				Fout1 = off;
			}
			
			// Toggle CLK
			Fout2 = (b%2)*hi;			

			// Update increments
			increment1 = (long) Fout1 * 68719L; // 68719000L ;
			increment2 = (long) Fout2 * 68719L; // 68719000L ;
	 		
			// pause for 25 mSec
			time = 25;
			count = countMS;
	 		while(time); 
		}
		
		// Ramp down
		ramp = 0;
		
		// Pause for 60 mSec
		time = 60;
		count = countMS;
	 	while(time); 
	}
	rel = trtCurrentTime() + SECONDS2TICKS(0.8;
	dead = trtCurrentTime() + SECONDS2TICKS(1.6);
	trtSleepUntil(rel, dead);
}

/****************************************************************************** 
* initialize()
*------------------------------------------------------------------------------
*    Initialize everything
*/
void initialize()
{
	// Turn on onboard LED
	DDRD = 0x04;
	PORTD = PORTD ^(0<<PORTD2);
	
	// Make B.3 an output, B.0 and B.2 inputs
	DDRB = (1<<PINB3) ;

	// Set up INT2 interrupt (pin B2) 
	EICRA = (1<<ISC21);     // set INT2 to interrupt on falling edge
	EIMSK = (1<<INT2);  	// enable interrupt by INT2
	
	// Initialize the sine and ramp tables
	for (i=0; i<256; i++)
	{
		sineTable[i] = (char)(63.0 * sin(6.283*((float)i)/256.0));
		rampTable[i] = i>>1;
	}

	// Set up Timer0
	TCCR0B = 1 ;            // set Timer0 to run at full rate
	TIMSK0 = (1<<TOIE0) ;   // turn on Timer0 overflow ISR

	// Set up PWM
	// turn on fast PWM and OC0A output
	// at full clock rate, toggle OC0A (pin B3) 
	// 16 microsec per PWM cycle sample time
	TCCR0A = (1<<COM0A0) | (1<<COM0A1) | (1<<WGM00) | (1<<WGM01) ; 
	OCR0A = 128 ; // set PWM to half full scale

	// Phase lock the sine generator DDS
	accumulator1 = 0;
	accumulator2 = 0;	

	// Crank up the ISRs
	sei();
}

/****************************************************************************** 
* main()
*------------------------------------------------------------------------------
*    Entry point and task scheduler loop
*/
int main(void)
{
	// Run initialization stuff
	initialize();

	// Initialize UART (trt_uart_init() is in trtUart.c)
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"\n\r Keyboard Reading Started...\n\r\n\r");

	// Start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack

	// Create Semaphores (first two are necessary for uart)
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; 	// uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  	// user typed <enter>
	trtCreateSemaphore(TRANS_BUFFER_RDY,0) ;  	// new key scan code in trans_buffer

	// Create tasks
	trtCreateTask(transmitter, 400, SECONDS2TICKS(0.1), SECONDS2TICKS(1),&(args[1]));

	// Initialize variables
	started = 0;
	shift = 0;
	extended = 0;
	kbd_scan_code = 0;
	bit_count = 0;
	parity = 0;
	transWR_pnt = 0;
	transRD_pnt = 0;
	count = 0;
	time = 0;
	ramp = 0;

	// Idle Task - just sleeps the cpu to save power every time it executes
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	while (1) {
		sleep_cpu();
	}
}


